#include <stdio.h> // printf
#include <stdlib.h> // calloc
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include <cutil.h> // CUDA error checking

// 2D to 1D bijection
#define IX(i,j) ((i)*(N)+(j))
// maximum
#define MAX(a,b) (((a)>(b))?(a):(b))

#ifndef N
#define N 1024
#endif

#ifndef B
#define B 32
#endif

#define SIZE (N*N)

__global__ void setmm(float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	const uint i = blockIdx.y*B + threadIdx.y;
	const uint j = blockIdx.x*B + threadIdx.x;
	a[IX(i,j)] = sinf((float)i*j);
	b[IX(i,j)] = cosf((float)i+j);
	c[IX(i,j)] = 0.0f;
}

__global__ void sgemm_shared(float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	__shared__ float sa[B][B];
	__shared__ float sb[B][B];

	const uint ti = threadIdx.y; const uint tj = threadIdx.x;
	const uint i = blockIdx.y*B + ti;
	const uint j = blockIdx.x*B + tj;
	float cij = 0.0f;
	for (uint m=0; m<N/B; ++m) {
		sa[ti][tj] = a[IX(i,m*B+tj)];
		sb[ti][tj] = b[IX(m*B+ti,j)];
		__syncthreads();

		for (uint k=0; k<B; ++k)
			cij += sa[ti][k] * sb[k][tj];
		__syncthreads();
	}
	c[IX(i,j)] = cij;
}

int main(int argc, char **argv)
{
	float *d_a = NULL, *h_a = NULL; // pointers to a, b and c in Host and Device
	float *d_b = NULL, *h_b = NULL;
	float *d_c = NULL, *h_c = NULL;
	assert(argc==1);
	assert(N%B==0); // evenly split

	CUDA_SAFE_CALL(cudaMalloc(&d_a, SIZE * sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc(&d_b, SIZE * sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc(&d_c, SIZE * sizeof(float)));
	h_a = (float *) calloc(SIZE, sizeof(float));
	h_b = (float *) calloc(SIZE, sizeof(float));
	h_c = (float *) calloc(SIZE, sizeof(float));
	assert(d_a && d_b && d_c && h_a && h_b && h_c);

	dim3 grid_size(N/B, N/B);
	dim3 block_size(B,B);
	setmm<<<grid_size, block_size>>>(d_a, d_b, d_c);
	CUT_CHECK_ERROR("setmm kernel failed");
	sgemm_shared<<<grid_size, block_size>>>(d_a, d_b, d_c);
	CUT_CHECK_ERROR("sgemm kernel failed");
	cudaDeviceSynchronize();

	CUDA_SAFE_CALL(cudaMemcpy(h_a, d_a, SIZE*sizeof(float), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_b, d_b, SIZE*sizeof(float), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_c, d_c, SIZE*sizeof(float), cudaMemcpyDefault));
	double max_diff = 0.0;
	for (unsigned int i=0; i<N; ++i) {
		for (unsigned int j=0; j<N; ++j) {
			float cij = 0.0f;
			for (unsigned int k=0; k<N; ++k)
				cij += h_a[IX(i,k)] * h_b[IX(k,j)];
			max_diff = MAX(max_diff, abs(cij-h_c[IX(i,j)]));
		}
	}
	printf("max_diff: %f\n", max_diff);

	free(h_c); free(h_b); free(h_a);
	CUDA_SAFE_CALL(cudaFree(d_c));
	CUDA_SAFE_CALL(cudaFree(d_b));
	CUDA_SAFE_CALL(cudaFree(d_a));

	return 0;
}
