#include <stdio.h> // printf
#include <stdlib.h> // calloc
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include <cutil.h> // CUDA error checking

// 2D to 1D bijection
#define IX(i,j) ((i)*(N)+(j))
// integer ceiling division
#define DIV_CEIL(a,b) (((a)+(b)-1)/(b))
// maximum
#define MAX(a,b) (((a)>(b))?(a):(b))

__global__ void setmm(const uint N, float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	const uint i = blockIdx.y*blockDim.y + threadIdx.y;
	const uint j = blockIdx.x*blockDim.x + threadIdx.x;
	if (i<N && j<N) {
		a[IX(i,j)] = sinf((float)i*j);
		b[IX(i,j)] = cosf((float)i+j);
		c[IX(i,j)] = 0.0f;
	}
}

__global__ void sgemm_unroll(const uint N, float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	const uint i = blockIdx.y*blockDim.y + threadIdx.y;
	const uint j = blockIdx.x*blockDim.x + threadIdx.x;
	if (i<N && j<N) {
		for (uint k=0; k<N; k+=4)
			c[IX(i,j)] += a[IX(i,k)] * b[IX(k,j)]
					   + a[IX(i,k+1)] * b[IX(k+1,j)]
					   + a[IX(i,k+2)] * b[IX(k+2,j)]
					   + a[IX(i,k+3)] * b[IX(k+3,j)];
	}
}

int main(int argc, char **argv)
{
	float *d_a = NULL, *h_a = NULL; // pointers to a, b and c in Host and Device
	float *d_b = NULL, *h_b = NULL;
	float *d_c = NULL, *h_c = NULL;
	assert(argc==4);
	const unsigned int N = atoi(argv[1]); // mtx side size
	const unsigned int BX = atoi(argv[2]); // block columns
	const unsigned int BY = atoi(argv[3]); // block rows
	const unsigned int SIZE = N*N; // linear size

	CUDA_SAFE_CALL(cudaMalloc(&d_a, SIZE * sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc(&d_b, SIZE * sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc(&d_c, SIZE * sizeof(float)));
	h_a = (float *) calloc(SIZE, sizeof(float));
	h_b = (float *) calloc(SIZE, sizeof(float));
	h_c = (float *) calloc(SIZE, sizeof(float));
	assert(d_a && d_b && d_c && h_a && h_b && h_c);

	dim3 grid_size(DIV_CEIL(N,BX), DIV_CEIL(N,BY));
	dim3 block_size(BX,BY);
	setmm<<<grid_size, block_size>>>(N, d_a, d_b, d_c);
	CUT_CHECK_ERROR("setmm kernel failed");
	sgemm_unroll<<<grid_size, block_size>>>(N, d_a, d_b, d_c);
	CUT_CHECK_ERROR("sgemm kernel failed");
	cudaDeviceSynchronize();

	CUDA_SAFE_CALL(cudaMemcpy(h_a, d_a, SIZE*sizeof(float), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_b, d_b, SIZE*sizeof(float), cudaMemcpyDefault));
	CUDA_SAFE_CALL(cudaMemcpy(h_c, d_c, SIZE*sizeof(float), cudaMemcpyDefault));
	double max_diff = 0.0;
	for (unsigned int i=0; i<N; ++i) {
		for (unsigned int j=0; j<N; ++j) {
			float cij = 0.0f;
			for (unsigned int k=0; k<N; ++k)
				cij += h_a[IX(i,k)] * h_b[IX(k,j)];
			max_diff = MAX(max_diff, abs(cij-h_c[IX(i,j)]));
		}
	}
	printf("max_diff: %f\n", max_diff);

	free(h_c); free(h_b); free(h_a);
	CUDA_SAFE_CALL(cudaFree(d_c));
	CUDA_SAFE_CALL(cudaFree(d_b));
	CUDA_SAFE_CALL(cudaFree(d_a));

	return 0;
}
